(* A type of integer-carrying trees *)
type tree = 
  | Empty
  | Node of tree * int * tree 


(* helper function for constructing trees *)
let leaf (i:int) : tree =
  Node(Empty, i, Empty)

(* determines whether a node is a leaf *)
let is_leaf (t:tree) : bool =
  begin match t with
    | Node(Empty,_,Empty) -> true
    | _ -> false
  end

(***********************)
(* tree pretty printer *)
(***********************)


(* a representation of pictures *)
type picture = string array

(* a picture with some extra information:
    rootx is the x coordinate (starting from 0)
    of the root character of this tree 
*)
type tree_block = {
  width : int;
  height : int;
  rootx : int;
  picture : picture;
}

let char_of_int (i:int) : char =
  Char.chr ((Char.code '0') + i)

(* creates a new tree block with a root value v 
   rooted at rx *)
let new_tb (w:int) (h:int) (rx:int) (v:int) =
  let pic = Array.init h (fun _ -> String.make w ' ') in
  let _ = String.set (pic.(0)) rx  (char_of_int v) in
    {width = w;
     height = h;
     rootx = rx;
     picture = pic;}

(* inserts newlines into a treeblock to serialize it
   as a single string *)
let string_of_tree_block (b:tree_block) : string =
  String.concat "\n" (Array.to_list (b.picture))

(* an exception that is raised if there is overlap 
   detected when trying to superimpose two trees *)
exception Overlap

(* copies one tree_block into another; doesn't check for overlap 
   x = starting x position
   dy = amount to shift the copied block vertically
*)
let overwrite (dst:tree_block) (src:tree_block) (x:int) (dy:int) : unit =
  for i = 0 to src.height - 1 do
    String.blit (src.picture.(i)) x (dst.picture.(i+dy)) x src.width
  done

(* copies on tree_block into another, checking for collisions (overlap)
   dx and dy are offsets to shift the source block relative to the target *)
let try_to_write (dst:tree_block) (src:tree_block) (dx:int) (dy:int) : unit =
  for i = 0 to src.height - 1 do
    for j = 0 to src.width - 1 do
      let y = i+dy in
      let x = j+dx in
      let src_char = String.get (src.picture.(i)) j in
	if    ((String.get (dst.picture.(y)) x) <> ' ') 
	   && (src_char <> ' ')
	then
	  raise Overlap
	else
	  String.set (dst.picture.(y)) x src_char
    done
  done

(* draws an left-slanting edge of length el starting at root
   rx *)
let draw_left_edge (dst:tree_block) (el:int) (rx:int) : unit =
  for i = 1 to el do
    String.set dst.picture.(i) (rx - i) '/'
  done

(* draws a right-slanting edge of length el starting at root
   rx *)
let draw_right_edge (dst:tree_block) (el:int) (rx:int) : unit =
  for i = 1 to el do
    String.set dst.picture.(i) (rx + i) '\\'
  done

(* superimposes two tree_blocks and gives them a new root r.
   it tries to connect the two trees to the root using an edge
   of length el.  Raises 'Overlap' if there is a collision in 
   this process *)
let superimpose (b1:tree_block) (b2:tree_block) (r:int) (el:int) : tree_block =
  let dx = el * 2 + 2 in
  let dy = el + 1 in 
  let new_height = dy + (max b1.height b2.height) in
  let (fixed_block, shifted_block, shift_amt, new_width, new_rootx) = 
    if (b1.rootx + dx) >= b2.rootx then
      (* b1 is not shifted *)
      let amt = b1.rootx + dx - b2.rootx in
      let w = max b1.width (b1.rootx + dx +(b2.width - b2.rootx)) in
      let rx = b1.rootx + dy in
	(b1, b2, amt, w, rx)
    else
      (* b1 is shifted *)
      let amt = b2.rootx - (b1.rootx + dx) in
      let w = max (b1.width + amt) b2.width in
      let rx = b1.rootx + amt + dy in
	(b2, b1, amt, w, rx)
  in
  let new_block = new_tb new_width new_height new_rootx r in
  let _ = overwrite new_block fixed_block 0 dy in
  let _ = try_to_write new_block shifted_block shift_amt dy in
  let _ = draw_left_edge new_block el new_rootx in
  let _ = draw_right_edge new_block el new_rootx in
    new_block

(* merges two tree blocks by trying to connect them to a new
   root.  It starts with the smallest possible edge length
   and increases it until no collision overlaps are detected *)
let merge (b1:tree_block) (b2:tree_block) (r:int) : tree_block =
  let rec loop (el:int) : tree_block =
    try superimpose b1 b2 r el with
      | Overlap -> loop (el+1)
  in
    loop 1


(* makes a tree_block into the left child of a new root r *)
let make_left_child (t:tree_block) (r:int) : tree_block =
  let new_height = t.height + 2 in
  let new_width = max (t.width) (t.rootx + 3) in
  let new_rootx = t.rootx + 2 in
  let new_block = new_tb new_width new_height new_rootx r in
  let _ = overwrite new_block t 0 2 in
  let _ = draw_left_edge new_block 1 new_rootx in
    new_block

(* makes a tree_block into the right child of a new root r *)
let make_right_child (t:tree_block) (r:int) : tree_block =
  let new_height = t.height + 2 in
  let shift_amt = max 0 (2 - t.rootx) in
  let new_width = t.width + shift_amt in
  let new_rootx = max 0 (t.rootx - 2) in
  let new_block = new_tb new_width new_height new_rootx r in
  let _ = try_to_write new_block t shift_amt 2 in
  let _ = draw_right_edge new_block 1 new_rootx in
    new_block

(* creates an empty tree_block *)
let empty_block () : tree_block =
  {height=0; width=0; rootx=0; picture=Array.make 0 ""}

(* creates an tree_block with just one node, which is 
   both the root and leaf *)
let leaf_block (r:int) : tree_block =
  {height=1; width=1; rootx=0; 
   picture=Array.make 1 (String.make 1 (char_of_int r));}

(* recursively constructs a tree_block from a tree
   using 'merge' to find an optimal non-overlapping 
   picture for the children *)
let rec block_of_tree (t:tree) : tree_block =
  begin match t with 
    | Empty -> empty_block ()
    | Node(Empty, r, Empty) -> leaf_block r
    | Node(Empty, r, rt) -> 
	let rtb = block_of_tree rt in
	  make_right_child rtb r
    | Node(lt, r, Empty) ->
	let ltb = block_of_tree lt in
	  make_left_child ltb r
    | Node(lt, r, rt) ->
	let ltb = block_of_tree lt in
	let rtb = block_of_tree rt in
	  merge ltb rtb r
  end

(* The main function:  
   assumes that the nodes are in [0-9]
*)
let string_of_tree (t:tree) : string =
  string_of_tree_block (block_of_tree t)

let top_print (ppf:Format.formatter) (t:tree) : unit =
	Format.fprintf ppf "\n%s" (string_of_tree t)